#! /usr/bin/perl 

#########################################################
#Xiangchao Gan 13-8-2014                                #
#BAMLINK perl interface script                          #
#This script;                                           #
#  --reads command options and check the parameters     #
#  
########################################################

use FindBin;
use lib "$FindBin::Bin/.";
use bamlink_common;

use Getopt::Long; 
use File::Spec::Functions qw(rel2abs);
use Scalar::Util qw(looks_like_number);
use File::Path;
use File::Basename;
use strict;


my $version = "v.0.1.0";

my $min_links_default;
my $max_link_ratio_default;
my $min_tig_overlap;
my $max_count_trim ;

my $outputfile;
my $help;
my $includechrs;
my $excludechrs;
my $base_name="blink"; 
my $doplot;
my $verbose;
my $tdir = "." ;
my $mapqual =20;

my $filecontig ;
my $libraryfile ;
my $project_config_file;

my $seplines = ("-" x 60)."\n";

my $bamlink_path = dirname(rel2abs($0));
GetOptions( 'config_file=s' => \$project_config_file, 
	    'out_tmpfolder=s' => \$tdir     ,
	    'out_tmpbasename=s'=> \$base_name  ,
	    'out_seqfile=s' => \$outputfile ,
	    'mapqual=s' => \$mapqual,
	    'verbose=s' => \$verbose,
	    'dotplot=s' => \$doplot,
	    'include=s' => \$includechrs,
	    'excludes=s' => \$excludechrs,
            'help!' => \$help
); 

if ($help || @ARGV<2 ) {
    help($version);
}

$filecontig=shift;
$libraryfile=shift;

#$includechrs="C[0-9]*";
#$excludechrs="ChrC; ChrM";

$includechrs =~ s/(\s)+/\1/g if ($includechrs);
$excludechrs =~ s/(\s)+/\1/g if ($excludechrs);
#print $includechrs, "\n";

&read_config_files($project_config_file);  # stores the configuration 
&check_parameters(); #checkin if user setted the parameters right


#----------------------------------------Setup folder structure------------------------
mkpath("$tdir/intermediate_results");
mkpath("$tdir/pairinfo");
mkpath("$tdir/reads");
my $dotfilepath="$tdir/dotfiles";
mkpath($dotfilepath );

#----------------------------------------Print default parameters and input sequence stats--------------------
my $subsample=defined $includechrs || defined $excludechrs ;
my $contig_excluded;
my $contig = "$tdir/intermediate_results/" . $base_name .  ".formattedcontigs.fasta";
if( !$subsample ){
    system( "cp $filecontig $contig" );
} else {
    $contig = "$tdir/intermediate_results/" . $base_name .  ".subcontigs.fasta";
    $contig_excluded="$tdir/" . $base_name . ".excludedcontigs.fasta";
    &fastaTransfer($filecontig, $contig, $contig_excluded, $includechrs, $excludechrs);
}

my $log = "$tdir/".$base_name . ".logfile.txt";
my $summaryfile = "$tdir/". $base_name.".summaryfile.txt";
my ($loghandle, $sumhandle);
open ($loghandle, ">$log") || die "Can't write to $log -- fatal\n";

open ($sumhandle, ">$summaryfile") || die "Can't open $summaryfile -- fatal\n";
close $sumhandle;

my $init_message =  "Your are running BAMLINK $version at ".getDate().":\nRequired inputs: \n\t--fasta_file=$filecontig\n\t--library_file= $libraryfile\n\n";
$init_message .= "Optional inputs:\n\t--out_tmpfolder = $tdir \n\t--out_tmpbasename = $base_name\n\tmapqual = $mapqual\n\tverbose = $verbose\n\tdoplot = $doplot\n\n";
&printMessage($loghandle, $init_message);
close $loghandle;


###UPDATE SUMMARY FILE
open ($sumhandle, ">>$summaryfile") || die "Can't open $summaryfile -- fatal\n";
open ($loghandle, ">>$log") || die "Can't write to $log -- fatal\n";

#write summary of initial contigs
my $sumfile .= "\nSUMMARY: \n".$seplines."\tcontig file;\n";
$sumfile .= &writesummaryfiles($filecontig, "contig");

if($subsample) {
    $sumfile .= "\tAfter subsampling contig file;\n";
    $sumfile .= &writesummaryfiles($contig, "contig");
}
&FlushFiles($loghandle, $sumhandle);
close $loghandle;
close $sumhandle;

#--------------------------------------------------READ BAM FILES 
open(FILELIB, "< $libraryfile") || die "Can't open $libraryfile -- fatal\n";
my ($lib, $baminputlist, $bamlibid, $insert_size, $insert_stdev, $pair, $headscaffolds, $prevlib, $mergedtigs, $evidencefile);

my %bam2sbam;
while(<FILELIB>){
    chomp;
    &FlushFiles($loghandle, $sumhandle);
    my ($lib, $data_type, $baminputlist, $bamlibid, $insert_size, $insert_stdev, $min_links2, $max_link_ratio2, $RFdistance) = split(/\s+/, $_);
    my ($min_links, $max_link_ratio, $fluctuation_allowed);

    next if ( $_ =~ /^\s*#/ );
    next if($lib eq $prevlib || $lib eq '');
    $prevlib = $lib;
     	if ( $min_links2) {
		$min_links=$min_links2;
	} else {
		$min_links=$min_links_default;	
	}
	if ( $max_link_ratio2 ) {
    		$max_link_ratio=$max_link_ratio2;
	} else {
    		$max_link_ratio=$max_link_ratio_default;
	}
    $fluctuation_allowed =  $insert_stdev * $insert_size;
    my $up_iz = ($insert_size + $fluctuation_allowed);

    open ($loghandle, ">>$log") || die "Can't write to $log -- fatal\n";
    &printMessage($loghandle, "\nLIBRARY $lib\n".$seplines);
    &printMessage($loghandle, "Parameters:\n");
    &printMessage($loghandle, "\tdata_type=$data_type\n\tmin_links = $min_links\n\tmax_link_ration = $max_link_ratio\n\tmin_contig_overlap = $min_tig_overlap\n\n" );
    close $loghandle;

    open ($sumhandle, ">>$summaryfile") || die "Can't open $summaryfile -- fatal\n";
    print $sumhandle "\n\nLIBRARY $lib STATS:\n".("#" x 80),"\n";
    close $sumhandle;

    my $scaffold = "$tdir/intermediate_results/" . $base_name . ".$lib.scaffolds";
    $mergedtigs = "$tdir/intermediate_results/" . $base_name . ".$lib.scaffolds.fasta";
    my $issues = "$tdir/pairinfo/" . $base_name . ".$lib.pairing_issues";
    my $distribution = "$tdir/pairinfo/" . $base_name . ".$lib.pairing_distribution.csv";


#-------------------------------------------------Scaffold the contigs and generate .scaffold file

    my @filess;
    if ( $baminputlist =~ /;/) {
	@filess= split(/;/, $baminputlist);
    } else {
        push(@filess, $baminputlist);	
    }
    my @newbams;
    foreach my $bfile (@filess) {
        my $sortbam;
        my $samtools_path="samtools";

	if (exists $bam2sbam{$bfile} ) {
           $sortbam = $bam2sbam{$bfile} ;
        } else {  
           open (BAMH, "$samtools_path view -H $bfile |") or die "Cannot open bam file: $bfile \n";
	   my $bamhead;
	   while ( <BAMH> ) {
		$bamhead=$_;
		if ($bamhead =~ /^[@]HD/ ) {
		    last;
		}
	   } 
	   close BAMH;

           my $is_namesorted= 0; 
	   if ($bamhead =~ /SO:queryname/) {
		$is_namesorted=1;
	   }
	
	   if ( $is_namesorted ) { 
                $sortbam = $bfile;
                $bam2sbam{ $bfile }=$sortbam; ###next time, do not need to read bam header again. 
	   } else {
	 	$sortbam="$tdir/reads/$base_name.$lib.nsort.bam";
		print "Name sorting file $bfile --> $sortbam\n";
		print "$samtools_path sort -n -o $sortbam $bfile", "\n";
		system("$samtools_path sort -n -o $sortbam $bfile");
                $bam2sbam{ $bfile }=$sortbam;  
	   }
        }   
        push(@newbams, $sortbam);
    }
    my $bams_str=join(",",@newbams);
    my $fileread = "$tdir/reads/$base_name.$lib.foundpairedreads.fasta";
    my $track_allfile = "$tdir/reads/$base_name.$lib.pairtrack.fasta";

    my $reverse=0;
    print("perl $bamlink_path/parsebam_paireseq.pl -t $tdir -b $base_name -s -q $mapqual  $contig $data_type $bams_str '$bamlibid' $fileread $track_allfile $insert_size $fluctuation_allowed $RFdistance\n");
    system("perl $bamlink_path/parsebam_paireseq.pl -t $tdir -b $base_name -s -q $mapqual  $contig $data_type $bams_str '$bamlibid' $fileread $track_allfile $insert_size $fluctuation_allowed $RFdistance");

    print "\n";
    print("perl $bamlink_path/pair2scaffold.pl -t $tdir -b $base_name $contig $fileread $track_allfile $issues $distribution $verbose $lib $insert_size $fluctuation_allowed $scaffold $min_links $max_link_ratio \n");
    system("perl $bamlink_path/pair2scaffold.pl -t $tdir -b $base_name $contig $fileread $track_allfile $issues $distribution $verbose $lib $insert_size $fluctuation_allowed $scaffold $min_links $max_link_ratio");
   

    &printMessage($loghandle, "\n=>".getDate().": Merging contigs and creating fasta file of scaffolds ($mergedtigs)\n");
    $evidencefile = $mergedtigs;
    $evidencefile =~ s/.fasta/.evidence/;
    print("perl $bamlink_path/mergcontig.pl $scaffold $contig $mergedtigs $evidencefile 50 $verbose $min_tig_overlap $max_count_trim\n"); 
    system("perl $bamlink_path/mergcontig.pl $scaffold $contig $mergedtigs $evidencefile 50 $verbose $min_tig_overlap $max_count_trim"); 

    
    $contig = $mergedtigs;
#-------------------------------------------------write summary of scaffolds
    $sumfile .= "\tAfter scaffolding $lib:\n";
    $sumfile .= &writesummaryfiles($mergedtigs, "scaffold");
    
#-------------------------------------------------
    open ($sumhandle, ">>$summaryfile") || die "Can't open $summaryfile -- fatal\n";
    print $sumhandle ("#" x 80),"\n";
    close $sumhandle;
    &printMessage($loghandle, "\n$seplines");

}#END OF LIBRARY LOOP
#-------------------------------------------------END OF LIBRARIES. PRINT SUMMARY TO FILE AND END SESSION


my $finalfile = "$tdir/$base_name" . ".final.scaffolds.fasta";
my $finalevfile = "$tdir/$base_name" . ".final.evidence";
open (EVID, $evidencefile);
open (FINALEV, "> $finalevfile");
while(<EVID>){
    print FINALEV $_;
}
close EVID;
close FINALEV;

open (SCAF, $mergedtigs);
open (FINAL, "> $finalfile");
while(<SCAF>){
    print FINAL $_;
}
close SCAF;
if ( $subsample) {
   open (REMA, "<$contig_excluded");
   while(<REMA>) {
      print FINAL $_;
   }
   close REMA;
}
close FINAL;


if (defined $outputfile) {
    system( "cp $finalfile $outputfile" );
    #copy( $finalfile, $outputfile) or die "copy failed for the final file. You can find the result at $finalfile instead\n";
}


#make .dot file for visualisation 
if($doplot) {
    &printMessage($loghandle,"\n=>".getDate().": Producing .dot file for visualisation\n");
    &visualiseScaffolds("$dotfilepath/$base_name.visual_scaffolds", $evidencefile);
}

open ($sumhandle, ">>$summaryfile") || die "Can't open $summaryfile -- fatal\n";
&printMessage($loghandle, "\n=>".getDate().": Creating summary file ($summaryfile)\n");
print $sumhandle $sumfile.$seplines;
&printMessage($loghandle, ("*" x 50)."\n\nProcess run succesfully on ".getDate()."\n\n\n");

close $loghandle;
close $sumhandle;
#END OF MAIN PROGRAM


sub read_config_files { #read config files in the form element = value #comment
  my $config_file = shift;
  my %parameters;
  my $fh_user_config;
  if (defined $config_file ) {
      open( $fh_user_config, "<", "$config_file") || die ("Couldn't open the project configuration file, you may have to check if the name or permission of the file are correct.\nDetails: $!\n");
  } else {    
      open( $fh_user_config, "<", "$bamlink_path/bamlinkdefault_config") || die ("The configuration file 'bamlinkdefault_config' couldn't be read, please check if the file exists and if its permissions are properly set.\n");
  }

  # -=-=-= PROJECT CONFIGURATION =-=-=-
  $min_links_default = read_config_file_line("min_links", $fh_user_config);
  $max_link_ratio_default = read_config_file_line("max_link_ratio", $fh_user_config);
  $min_tig_overlap  =  read_config_file_line("min_contig_overlap", $fh_user_config);
  $max_count_trim = read_config_file_line('max_count_trim',  $fh_user_config);
  $verbose = read_config_file_line('verbose',  $fh_user_config);
  $doplot = read_config_file_line('doplot',  $fh_user_config); 
}



sub read_config_file_line { # file format element = value
  my ($parameter, $fh_config_file) = @_;

  seek($fh_config_file, 0, 0);              # added to allow any order of parameters in the config files, preventing unfriendly error messages if the user changes the order
  while (my $line = <$fh_config_file>){
    if ($line =~ /^\s*$parameter\s+/) {    # the string to be searched in the file
      chomp ($line);
      $line =~ s/^\s*$parameter\s+//;   # removing what comes before the user input
      $line =~ s/#.*$//;                    # removing what comes after the user input (commentaries)
      $line =~ s/\s*$//;                    # removing what comes after the user input (space caracteres)
      if ($line eq 'undef' || $line eq '') { return; }
      else { return $line; }
    }
  }
  return;
}


sub check_parameters { #check for all parameters, 
  if(! $libraryfile || ! $filecontig ){
      help( $version );
      die();
  }
  die "ERROR: Invalid contig file $filecontig ...Exiting.\n" if(! -e $filecontig);
  die "ERROR: Invalid library file $libraryfile ...Exiting.\n" if(! -e $libraryfile);
  die "ERROR: --min_links must be an integer number. Your value is $min_links_default ...Exiting.\n" if(!($min_links_default * 1 eq $min_links_default));
  die "ERROR: --min_contig_overlap must be an integer number. Your value is $min_tig_overlap ...Exiting.\n" if(!($min_tig_overlap * 1 eq $min_tig_overlap));


  #-------------------------------------------------check library file;
  open(FILELIB, "< $libraryfile");
  my $countline=0;
  while(<FILELIB>){
    chomp;
    $countline++;
    next if ( $_ =~ /^\s*#/ );
    my @line = split(/\s+/, $_);
    if($#line >= 0){
	die "ERROR: Line $countline in your library file ($libraryfile) contains $#line spaces, which should be >=5 spaces. Check that no spaces are within the file names.\n" if($#line < 5);

        my ($library, $data_type, $baminputlist, $bamlibid, $insert_size, $insert_stdev, $min_links, $max_link_ratio, $RFdistance) = split(/\s+/, $_);

        if ( $library =~ /^bam/) {
    	    $baminputlist=$baminputlist;
    	    $bamlibid=$bamlibid;
        } else {
            die "Only support bam file as input";
        }
    
        my @filess;
        if ( $baminputlist =~ /;/) {
    	    @filess= split(/;/, $baminputlist);
        } else {
            push(@filess, $baminputlist);	
        }
        foreach my $bfile (@filess) {
            if (! -e $bfile) {
    	        die("ERROR: bam file $bfile cannnot been found\n");
    	    }
        }
    
	die "ERROR: data_type [col 4] must be PAIRENDED/MATEPAIR/PACBIO. Your value for library $library is $data_type ...Exiting.\n" if( $data_type !~ /PAIRENDED/i && $data_type !~ /MATEPAIR/i && $data_type !~ /PACBIO/i );
	die "ERROR: insert_size [col 5] must be an integer number. Your value for library $library is $insert_size ...Exiting.\n" if( !looks_like_number($insert_size));
	die "ERROR: insert_size [col 5] should be higher than or equal to 0. Your value for library $library is $insert_size. Exiting.\n" if(!($insert_size>0) || !looks_like_number($insert_size) );
	die "ERROR: Insert stdev [col 6] must be a number between 0.00 and 1.00. Your value for library $library is $insert_stdev. Exiting.\n" if($insert_stdev < 0 || $insert_stdev > 1 || !looks_like_number($insert_stdev) );
	die "ERROR: min_links[col 7] must be an interger number. Your value for library $library is $min_links. ...Exiting.\n" if ($min_links && !looks_like_number($min_links));
	die "ERROR: max_link_ratio [col 8] must be a number between 0.00 and 1.00. Your value for library $library is $max_link_ratio ...Exiting.\n" if($max_link_ratio && (!looks_like_number($max_link_ratio) ||  $max_link_ratio < 0 ||$max_link_ratio > 1));
	die "ERROR: RFdistance [col 9] must be an integer number. Your value for library $library is $RFdistance ...Exiting.\n" if($RFdistance && !looks_like_number($RFdistance));
      }
  }
  close FILELIB;
}


sub help {
  my $ver = $_[0];
  print "\nBAMLINK $ver\n\n";

  print "Usage: $0 [options] input.fasta data_description_file\n\n";
  print "--config_file   config file for the default parameters (default bamlinkdefault_config, optional)\n";
  print "--out_tmpfolder path to the output folder (default: './')\n";
  print "--out_tmpbasename  Base name for all output files (default: 'blink')\n";
  print "--out_seqfile   output sequence file in fasta format (default: undefine)\n";
  print "--mapqual       Only use the read pairs with the mapping quality of both ends >= threshold (default: 20)\n"; 
  print "--exclude       Exclude contigs matching PATTERN\n";
  print "--include       Don't exclude contigs matching PATTERN\n";
  print "--verbose       Runs in verbose mode (-v 1=yes, -v 0=no, default: 0)\n";
  print "--dotplot       Make .dot file for visualisation of scaffolding (-p 1=yes, -p 0=no, default: 0)\n";
  die "\n";
}


