#! /usr/bin/perl

#########################################################
#Xiangchao Gan      13-8-2014                           #
#important fileprocessing functions for BAMLINK         # 
#                                                       #
#########################################################
#
#
package bamlink_common ;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(fastaTransfer writesummaryfiles printMessage getDate CounterPrint FlushFiles reverseComplement visualiseScaffolds readFileContigHashEx); 

use FindBin qw($Bin);
use lib "$Bin/dotlib/";
use DotLib;
use feature 'state';
use strict;

sub isIn {
    my ($name, $incl, $excl) = @_;
    my @in=split(/[; ]/, $incl);
    my @out=split(/[; ]/, $excl);	    
    foreach my $k (@in) {
	if ($name =~ /^$k$/) {
            return 1;
	}
    }
    foreach my $k (@out) {
        if ($name =~/^$k$/) {
            return 0;		
	}
    }
    return 1;
}



sub next_sequence {
    state $prevhile;
    state $prevhead;
    
    my $refh=shift;
    if (eof($refh)) {
	return 0;    
    }

    if ($refh != $prevhile ) {
        $prevhead="";
    }

    my $rseq;
    my $seq="";
    while (<$refh>) {
        s/\r\n/\n/;
        chomp;
        if ( /^\>(\S+)/ || eof($refh)) {
            my $head=$1;
            if($prevhead ne ''){ 
		$rseq->{id}=$prevhead;
		if ( eof($refh) && $_ !~ /^\>(\S+)/ ) {
                    $seq .=$_;   
		}  
		$rseq->{seq}=$seq;
		$prevhile=$refh;
		$prevhead=$head;
		return $rseq;
            }
	    $prevhead=$head;
       } else {
            $seq .= $_;
       }
    }
}


sub fastaTransfer {
    my ( $file, $out, $out_excluded,  $incl, $excl) = @_;
    if ( (!$incl && !$excl) || ($incl eq "" && $excl eq "") ) {
        system( "cp $file $out" );
	return;
    }

    my $countContig = 0;
    open(my $infasta,$file) || die "Can't open $file -- fatal\n";
    open(OUT, ">$out") or die "Can't open $out --fatal\n";
    
    my $out_ex;
    if ( $out_excluded ne "") {
        open($out_ex, ">$out_excluded") or die "Can't open $out_excluded --fatal\n";
    }    
     
    while ( my $p = next_sequence( $infasta ) ) {
                CounterPrint(++$countContig);
                if ( isIn($p->{id}, $incl, $excl) ) { 
		    print OUT ">", $p->{id}, "\n";
		    print OUT $p->{seq}, "\n";
	        } elsif ($out_excluded ne "") {
		    print "\nexlclude contig: $p->{id} \n";
		    print $out_ex ">",$p->{id}, "\n";
		    print $out_ex $p->{seq}, "\n";
		}
    }
    close $infasta;
    close OUT;
    close $out_ex if ( $out_excluded ne "");
}


###WRITE SUMMARY STATISTICS FOR ALL CONTIGS OR SCAFFOLDS
sub writesummaryfiles{
    my ($input_file, $insert) = @_;

    open (my $inhandle, $input_file) || die "Can't open input file $input_file.\n";

    my ($counter, $sum, $seq, $name, $foundN50, $sumN50) = (0,0, "","", 0, 0);
    my (@line, @lengths);

    while ( my $p = next_sequence( $inhandle ) ) {
	my $seq=$p->{seq};
	push(@lengths, length($p->{seq}));
	$sum+= length($p->{seq});
	$counter++;
    }
    my $half_length = $sum/2;
    
    my @lengths2 = reverse sort { $a <=> $b } @lengths;
    
    for(my $i = 0; $i <= $#lengths && $foundN50 == 0; $i++)
    {
	$sumN50 += @lengths2[$i];
	if($sumN50 >= $half_length){
	    $foundN50 = @lengths2[$i] if($sumN50 >= $half_length);
	    last;
	}
    }
    my $sumfile .= "\t\tTotal number of $insert"."s = $counter\n";
    $sumfile .= "\t\tSum (bp) = ". $sum. "\n";
    $sumfile .= "\t\tMax $insert size = ". @lengths2[0]."\n";
    $sumfile .= "\t\tMin $insert size = ". @lengths2[$#lengths]."\n";
    $sumfile .= "\t\tAverage $insert size = ".int($sum/$counter)."\n";
    $sumfile .= "\t\tN50 = ". $foundN50. "\n\n";
    
    close $inhandle;
    close OUTFILE;

    return $sumfile;
}



#READ THE CONTIG TO A HASH AND STORE THIS HASH
sub readFileContigHashEx{
    my ($file) = @_;
    
    my ($contigs, $tig_length, $chr2id);
    my $countContig = 0;
    open(my $infasta,$file) || die "Can't open $file -- fatal\n";

    while ( my $p = next_sequence( $infasta ) ) {
	    #print $p->{id}, "\n";
		CounterPrint(++$countContig);
		$tig_length->{$countContig} = length($p->{seq});
		$contigs->{$countContig}{'name'} = $p->{id};
		$contigs->{$countContig}{'seq'} = $p->{seq};
		$chr2id->{$p->{id}}=$countContig;
    }
    CounterPrint("                ");
    &FlushFiles();
    close $infasta;
    return ($contigs, $tig_length, $chr2id );
}


###############################################
#Functions below is copied from SSPACE        #
#                                             #  
###############################################

###FUNCTION TO REVERSE COMPLEMENT A SEQUENCE
sub reverseComplement{
    $_ = shift;
    tr/ATGC/TACG/;
    return (reverse());
}

###FUNCTION TO PRINT MESSAGES TO THE SCREEN AND TO THE LOG FILE
sub printMessage{
    my $hf=shift;	
    my $message = shift;
    print $message;
    print {$hf} $message;
}

###FUNCTION TO GET THE CURRENT DATE
sub getDate{
    my $date = scalar(localtime);
    return $date;
}

###PRINTS A COUNTER ON THE SCREEN AND OVERWRITES PREVIOUS LINE
sub CounterPrint{
    my $countingMessager = shift;
    print "\r$countingMessager";
    $|++;
}

###FLUSHES THE SUMMARY AND LOG FILE
sub FlushFiles{ 
    my $k=scalar (@_);	
    for (my $i=0; $i< $k; $i++) { 	
       my $handle=shift;	    
       select((select($handle), $| = 1)[0]);
    }
    $|++;
}


###FUNCTION TO GENERATE A VISUALISATION OF THE SCAFFOLDS AND THEIR CONTIGS IN .DOT FORMAT
sub visualiseScaffolds{
    my ($dotname, $evidence) = @_;
    #my ($filext, $sizecutoff) = (1, 5000000);
    my ($filext, $sizecutoff) = (1, 50000);
    my $filename2 = "$dotname.part".$filext.".dot";

    open(IN,$evidence) || die "can't read $evidence -- fatal\n";
    open(DOT, ">$filename2") || die "can't open $filename2 -- fatal\n";
    printHeader(\*DOT, undef);
    my ($prevtig, $prevgap, $prevlinks, $prevratio, $scafcount) = ("","","", "",0);
    while(<IN>){
	chomp;
	my $line = $_;
	my $filesize = -s $filename2;

	if ($line =~ /^[>]/){
	    endCluster(\*DOT) if($scafcount > 0);
	    my $filesize = -s $filename2;
	    if($filesize > $sizecutoff){
		printFooter(\*DOT);
		close(DOT);
		$filext++;
		$filename2 = "$dotname.part".$filext.".dot";
		open(DOT, ">$filename2") || die "can't open $filename2 -- fatal\n";
		printHeader(\*DOT, undef);
	    }
	    $scafcount++;
	    $line =~ tr/[>\|]/ /;
	    startCluster(\*DOT, $scafcount, "$line");
	    ($prevtig, $prevgap, $prevlinks, $prevratio) = ("","","", "");
	}
	elsif($line =~ /^[fr]/){
	    my @info = split(/\|/, $line);
	    my ($tnum, $sizetig, $links, $gap);
	    if($#info eq 1){
		($tnum, $sizetig) = split(/\|/, $line);
	    }else{
		($tnum, $sizetig, $links, $gap) = split(/\|/, $line);
	    }
	    my ($orient, $tig) = split(/_/,$tnum);
	    my $ori=-1;
	    my ($other, $gap2) = split(/gaps/,$gap);
	    my ($other, $links2) = split(/links/,$links);
	    $ori = 1 if($orient eq "f");
	    printNode(\*DOT, $tig, "$tig ($sizetig)", $ori);
	    printEdge(\*DOT, $prevtig, $tig, "gap = $prevgap links = $prevlinks", undef) if($prevtig ne "");

	    $prevtig = $tig;
	    $prevgap = $gap2;
	    $prevlinks = $links2;
	}
    }
    endCluster(\*DOT) if($scafcount > 0);
    printFooter(\*DOT);
    close(DOT);
    close IN;
}





#########END OF THE SCRIPT
