#!/usr/bin/perl

###################################################
#Xiangchao Gan 13-8-2014                          #
#BAMLINK perl subscript mergecontig.pl            #
#revised from SSPACE file mergecontig.pl          #
#This script;                                     #
#  -merge the contigs                             #
###################################################
#
use FindBin;
use lib "$FindBin::Bin/.";
use bamlink_common;

my $scaffold=shift;
my $contigs_file=shift;

my ($contigs) = &readFileContigHashEx($contigs_file);
mergeContigs($scaffold, $contigs, @ARGV);


###MAKE A .FASTA FILE OF THE FOUND SCAFFOLDS. EITHER MERGE TWO CONTIGS WHEN A OVERLAP OF -n EXISTS OR PLACE A GAP
sub mergeContigs{

    my ($scaffold, $contigs, $mergedtigs_file, $evidence_file, $chunk, $verbose,$min_tig_overlap,$max_count_trim, $nameprefix) = @_;
   
    if (!$nameprefix) {
       $nameprefix="NSCAFD.";
    }
    print "Use '$nameprefix"."XX' as the the name of new scaffolds\n";

    my %namemap;
    for my $ztig (keys %$contigs) {
	$namemap{ $contigs->{$ztig}{'name'} }= 1;
    }


    open(IN,$scaffold) || die "can't read $scaffold -- fatal\n";

    open(SCAFS,">$evidence_file") || die "can't write to $evidence_file -- fatal\n";
    open(OUT,">$mergedtigs_file") || die "can't write to $mergedtigs_file -- fatal\n";
    my $scafhashcount = keys ( %$headscaffolds );
    my $scaffoldHashStart;
    my ($tot,$sct,$ct_merge) = (0,0,0);
    while(<IN>){### each line is a scaffold
	chomp;
	my $sc="";;
	my @a = split(/\,/);
	my @tig;

	if($a[2]=~/\_/){
	    @tig = split(/\_/,$a[2]);
	}else{
	    push @tig, $a[2];
	}
	CounterPrint(++$sct);
	#$sct++;
        if ($sct==280) {
	    my $tt=1;
	}
	
	my ($ct,$tigsum,$mct) = (0,0,0);
	my ($prev,$word,$template) = ("NA","NA","NA");
	my ($seq,$prevseq,$headconcat,$prevEstimatedDistance, $prevLinks) = ("","","","");
	my $prev_tnum=-1;   ##keep the original contig name if one contig has not been used in scaffoling.
	foreach my $t (@tig){### each contig
	    $ct++;
	    
	    if($t=~/([fr])(\d+)z(\d+)(\S+)?/i){

		my $orient = $1;
		my $tnum=$2;
		$prev_tnum=$tnum;
		my $head = $orient . $tnum;
		my $search = "tig" . $tnum;
		my $other = $4;
		$tot+= $3;
		$tigsum +=$3;

		my ($estimatedDistance, $links) = ("", "");
		$estimatedDistance = $1 if($other=~/m((\-)?\d+)/);
		$links = $1 if($other=~/k((\-)?\d+)/);
		print "\tSC $a[0] - TIG $ct.  pattern: $t search: $search totalTigSize: $tot Orientation: $orient Gap/Overlap estimated distance: $estimatedDistance\n" if($verbose);

		my $count_trim = 0;

		$seq = $contigs->{$tnum}{'seq'};
		$seq = reverseComplement($seq) if($orient eq "r");
		chomp $seq;
		my $prev;
		if($scafhashcount >0){
		    $prev = $headscaffolds->{$tnum}{'head'};
		    $prev =~ s/^\n//;
		    chomp $prev;
		    delete $headscaffolds->{$tnum};
		    chomp $prev;
		    if($orient eq "r"){
			my @prevarray = split("\n", $prev);
			if($#prevarray >=0){
			    my $newprev="";
			    my ($tnum, $sizetig, $links, $gap, $prevline) = ("","","","","");
			    for(my $i = $#prevarray; $i >= 0; $i--){

				my @info = split(/\|/, $prevarray[$i]);
				if($#info eq 1){
				    ($tnum, $sizetig) = split(/\|/, $prevarray[$i]);
				}else{
				    ($tnum, $sizetig, $links, $gap) = split(/\|/, $prevarray[$i]);
				}
				$tnum =~ tr/fr/rf/;
				$newprev .= $prevline."|".$links."|".$gap."\n" if($prevline ne "");
				$prevline = $tnum."|".$sizetig;
			    }
			    $newprev .= $prevline;
			    $prev = $newprev;
			}
		    }
		}
		else{
		    $prev = "$orient"."_$search|size".length($seq);
		}
		$prev .= "|links$links|gaps$estimatedDistance" if($links ne "");

		$prev .= " (". $contigs->{$tnum}{'name'} . ")";


		#print "$prev\n";
		if($word ne "NA"){
		    #####
		    if(length($seq)<=$chunk){
			$template = $seq;
		    }else{
			$template = substr($seq,0,$chunk);
		    }

		    ##### word search
		    my $dynamic_word = $word;
		    if($prevEstimatedDistance <= 0){
		      SCAN:
			until($template =~ /$dynamic_word/){
			    $dynamic_word = substr($dynamic_word,1,length($dynamic_word));
			    if(length($dynamic_word) < $min_tig_overlap){
				$count_trim++;
				last SCAN if($count_trim >= $max_count_trim);
				$dynamic_word = substr($word,0,length($word)-$count_trim);
			    }
			}
		    }
		    if($prevEstimatedDistance <= 0  && $seq =~ /^\S{0,$max_count_trim}$dynamic_word(.*)/){### will grab the left-most match which is ok
			my $tail = $1;
			my $all = "ERROR_";
			while($prevseq =~ /^(.*)$dynamic_word/ig){
			    $all = $1;
			}
			print "$prevseq **** $all **** WORD:$word *** DWord:$dynamic_word *** COUNTTRIM:$count_trim\n" if($all=~/ERROR/);

			$prevseq = $all . lc($dynamic_word) . $tail;
			my $overlap = length($dynamic_word);
			$ct_merge++; 
			print "$ct_merge. GROUNDS FOR MERGING ($overlap nt overlap) !!!\n" if($verbose);
			$headconcat .= "|merged$overlap"."\n".$prev;
		    }else{
			### ADDED RLW 5.MAR.2010
			if($prevEstimatedDistance <= 0){
			    $prevseq .= "n" . $seq
			}else{
			    $prevseq .= ("N" x $prevEstimatedDistance) . $seq;
		    }
			$headconcat .= "\n".$prev;

		    }
		}else{
		    $prevseq = $seq;
		    $headconcat = "\n".$prev;
		    $mct++;
		}

		##### For the next search
		if(length($seq)<=$chunk){
		    $word = $seq;
		}else{
		    $word = substr($seq,length($seq)-$chunk,$chunk); ### this will be the next word to search with
		}
		$prevEstimatedDistance = $estimatedDistance;
		$prevLinks = $links;
	    }#tig regex

	}#each tig
	my $scsz = length($prevseq);
	$scaffoldHashStart->{$sct}{'head'} = $headconcat;

	my @line = split(/\n/, $headconcat);
	my $fa_t_id;
	if ( $#line >1 ) {
##	    $fa_t_id="$a[0].$mct|size$scsz";
	    $fa_t_id="$nameprefix"."$sct";
	    while ( $namemap{ $fa_t_id} ) {
		$fa_t_id .= "I";
	    }
	    print SCAFS ">$fa_t_id"."|tigs".($#line)."$headconcat\n\n";
	} else {
	    $fa_t_id=$contigs->{$prev_tnum}{'name'};
	}
	print OUT ">$fa_t_id\n$prevseq\n";

	$prevseq = '';
    }
    close IN;
    close SCAFS;
    close OUT;
    CounterPrint("                ");
    undef $contigs;
    &FlushFiles();
    return ($scaffoldHashStart, $evidence_file);
}
