#! /usr/bin/perl



###################################################
#Xiangchao Gan 13-8-2014                          #
#BAMLINK perl subscript bam_pairseqparse.pl       #
#This script;                                     #
#  -reads the contig sequences in a hash          #
#  -stores sequence info in a hash                #
###################################################

use strict;
use FindBin;
use lib "$FindBin::Bin/.";
use bamlink_common;

use Storable;
use File::Path;
use File::Basename;
use List::Util qw(first max maxstr min minstr reduce shuffle sum);

use Getopt::Std;
my %opts = (t=>'.', b=>"blink", q=>20, m=>45, s=>undef);
getopts('t:b:q:s', \%opts);
my $tdir =$opts{t};
my $base_name = $opts{b};
my $is_namesorted = $opts{s};
my $thr_mapqual=$opts{q};
my $thr_minmaplen=$opts{m};

my $contigs_file= $ARGV[0];
my $data_type= $ARGV[1];
my $baminputfile =$ARGV[2];
my $libname=$ARGV[3];
my $fileread=$ARGV[4];
my $track_allfile= $ARGV[5];
my $insert_size = $ARGV[6];
my $fluctuation_allowed = $ARGV[7];
my $RFdistance=$ARGV[8];
if (! defined $RFdistance) {
    $RFdistance=0;
}

my $seplines = ("-" x 60)."\n";

my $reverse=0;
if ( $data_type =~/MATEPAIR/i ) {
    $reverse=1;
}

my ($low_iz, $up_iz) = ($insert_size - $fluctuation_allowed, $insert_size + $fluctuation_allowed);
my $log = "$tdir/$base_name" . ".logfile.txt";
my $summaryfile = "$tdir/$base_name" .".summaryfile.txt";

my ($loghandle, $sumhandle);
open ($loghandle, ">>$log") || die "Can't write to $log -- fatal\n";
open ($sumhandle, ">>$summaryfile") || die "Can't open $summaryfile -- fatal\n";

my($contighandle, $tig_length, $chr2idmap) = &readFileContigHashEx($contigs_file);


my ($cn_ok_contig,$cn_pcrdup_contig, $cn_lowmapq_contig, $cn_wrongori_contig, $cn_outbound_contig)=(0,0,0,0,0);
my $cn_lowmapq_pairs=0;
my $cn_deepinside_pairs=0;
my $cn_scaffolding_pairs=0;
my $cn_nomap=0;  #at least one end unmapped

my ($counter, $step) = (0, 100000);
&parseBam($baminputfile, $libname, $data_type, $chr2idmap, $fileread, $track_allfile, $thr_mapqual);
undef $chr2idmap;



close $sumhandle;
close $loghandle;



#-------------------------------------------------
###READ BAM FILE AND STORE IT IN A HASH
sub parseBam{
    my ($filestr, $libname, $data_type, $chr2id,  $filefoundread, $filetrack, $thr_mapq) = @_;

    ##my $track_all;
    &printMessage($loghandle,"\n=>".getDate().": Storing alignment output ($filestr) to memory\n");

    open OUTFOUND, "> $filefoundread";

    open OUTTRACK, "> $filetrack";


    my $samtools_path="samtools";
    my @filess;	
    if ( $filestr =~ /;/) {
	@filess= split(/;/, $filestr);
    } else {
	push(@filess, $filestr);
    }
    for my $file (@filess)  {
	if ( $file =~ /.bam$/ ) {
	    my $sortbam;
	    if ( defined $is_namesorted) {
		$sortbam = $file;
	    } else {
		$sortbam="$tdir/reads/$base_name.sortbam.bam";
		system("$samtools_path sort -n -o $sortbam $file");
            }

            if ($libname eq "*") { 
	        open (IN,"$samtools_path view $sortbam |") or die "Cannot open bam file: $sortbam $!\n";
	    } else {
	        open (IN,"$samtools_path view -l \"$libname\" $sortbam |") or die "Cannot open bam file: $sortbam $!\n";
	    }
	} else {
	    open(IN,$file) || die "Can't open $file -- fatal\n";
	}


        my $pre_id="";
	my @sticky;
	while(<IN>){
	    next if /^[@]/;
	    chomp;
	    ++$counter;
	    if($counter == $step){
		CounterPrint($counter);
		$step = $step + 100000;
	    }

	    my @bam_t= split(/\t/,$_);
	    my ($read, $flag, $contig, $start, $mapq, $cigar,$rnext, $pnext, $TLEN, $seq,$other) = @bam_t;
	    next if ($flag & 0x100);

            if ( ($pre_id && $pre_id ne $bam_t[0]) ) {
		&process_sticky(\@sticky, $chr2id, $thr_mapq);
		$pre_id = $bam_t[0];
            } elsif ( ! $pre_id ) {
		$pre_id = $bam_t[0];
            }
            push(@sticky, \@bam_t);
	}
	if (@sticky >0 ) {
	    &process_sticky(\@sticky, $chr2id, $thr_mapq);
	}
	close IN;
    }

    close OUTFOUND;
    close OUTTRACK;
    CounterPrint("                ");
    print $sumhandle "\nMAPPING READS TO CONTIGS:\n";
    print $sumhandle "\t\tAt least one end is PCR duplate or fails flatform QC: $cn_pcrdup_contig\n";
    print $sumhandle "\t\tAt least one end is unmappable: $cn_nomap\n";

    print $sumhandle "\n\tPair within a single contig (pre-scaffold):\n";
    print $sumhandle "\t\tAt least one end is of low mapping quality: $cn_lowmapq_contig\n";
    print $sumhandle "\t\tUnsatisfied orientation(i.e. ->->, <-<- but allowing for <--> ): $cn_wrongori_contig\n";
    print $sumhandle "\t\tUnsatisfied in distance(i.e. calculated distances out-of-bounds $insert_size +/- $fluctuation_allowed): $cn_outbound_contig\n";
    print $sumhandle "\t\tSatisfied in distance/logic: $cn_ok_contig\n";

    print $sumhandle "\n\tPair within two contigs:\n";
    print $sumhandle "\t\tAt least one end is of low mapping quality: $cn_lowmapq_pairs\n";
    print $sumhandle "\t\tAt least one end is unusable, either alignment seq is too short, or too deep inside(gap-filling instead): $cn_deepinside_pairs\n";
    print $sumhandle "\t\tNumber of pairs within two contigs for further analysis = ". $cn_scaffolding_pairs ."\n"; 
    print $sumhandle "$seplines\n";
    &FlushFiles($loghandle, $sumhandle);
    return $filetrack, $filefoundread;
}

sub process_sticky {
    my ($tik, $chr2id, $thr_mapq) =@_;

    if (@$tik != 2 ) {
        @$tik=();
        return;
    }
    my ($read, $flag, $contig, $start, $mapq, $cigar,$rnext, $pnext, $TLEN, $seq,$other) = @{$tik->[0]};
    my ($read2, $flag2, $contig2, $start2, $mapq2, $cigar2,$rnext2, $pnext2, $TLEN2, $seq2,$other2) = @{$tik->[1]};
    
    if ( $contig eq "*" || $contig2 eq "*" ) {
	$cn_nomap++;
	@$tik=();
	return;
    }
    if ( $flag & 0x200 || $flag & 0x400 || $flag2 & 0x200 || $flag2 & 0x400) {
	$cn_pcrdup_contig++;
	@$tik=();
	return;
    } 

    if ($rnext eq "=" || $rnext eq $contig ) {
	if ( $mapq <$thr_mapq || $mapq2 <$thr_mapq) {
	    $cn_lowmapq_contig++;
	    @$tik=();
	    return;
	} elsif ( ($flag & 0x10) == ($flag2 & 0x10) ) {
	    $cn_wrongori_contig++;
	    @$tik=();
	    return;
	} elsif ( max($start+length($seq), $start2+length($seq2))- min($start, $start2)<$low_iz || max($start+length($seq), $start2+length($seq2))- min($start, $start2) >$up_iz  ) {
	    $cn_outbound_contig++;
	    @$tik=();
	    return;
	} else {
	    $cn_ok_contig++;
	    @$tik=();
	    return;
	}
    }

    if ( $mapq <$thr_mapq || $mapq2 <$thr_mapq) {
	$cn_lowmapq_pairs++;
	@$tik=();
	return;
    }

    my ($seq1, $keyvalue1) = StoreResults( $tik->[0], $chr2id, $reverse);
    my ($seq2, $keyvalue2) = StoreResults( $tik->[1], $chr2id, $reverse);
    if ( !$keyvalue1 || !$keyvalue2) {
	$cn_deepinside_pairs++; 
    } elsif ($seq1 && $seq2) {
	$cn_scaffolding_pairs++;
	print OUTFOUND ">$counter\n$seq1:$seq2\n";
	print OUTTRACK "$seq1:$keyvalue1";
	print OUTTRACK "$seq2:$keyvalue2";
    }
    @$tik=();
    return
}

sub StoreResults{
    my ($input, $chr2id,$reverse) = @_;
    my ($read, $flag, $rname, $start, $mapq, $cigar, $rnext, $pnext,$TLEN, $seq, $other) = @$input;

    my ($tig,$startval, $endval);
    $start--;     ###start with 0;
    my $strand="+";
    if ( $flag & 0x10 ) {
	$strand="-";
    }
    $tig=$chr2id->{$rname};
    if (! $tig) {     ###allow hierarchical scaffolding, the one has already been scaffolding before won't be used any more.
	return;
    }
    my $keyvalue;	

    my $lower = ($up_iz+200);
    my $upper;
    $upper = ($tig_length->{$tig} - ($up_iz + 100)) if($tig_length->{$tig} > ($up_iz +100));
    $upper = $lower if($tig_length->{$tig} <= ($up_iz +100));

    my $RFlower=$RFdistance;
    my $RFupper= ($tig_length->{$tig} - $RFdistance ) ;

    my $seqlen=length($seq);
    my $lefthang=0;
    my $righthang=0;
    $cigar =~ /^([0-9]+)S/; $lefthang=$1;
    $cigar =~ /([0-9]+)S$/; $righthang=$1;
    if ($lefthang + $righthang>0) {
        $seqlen=$seqlen - $lefthang- $righthang;
	$seq = substr($seq, $lefthang, $seqlen);
	if ($seqlen < $thr_minmaplen) {
	    return;
	}
    }

    if($start <= ($lower) || $start >= ($upper)){
	if ( $RFlower <=0 || ($start >=$RFlower &&  $start<= $RFupper) )  {
	    ###### $seq is the reverse complement for the reverse library (long insert size)
	    if( ($strand eq "+" && !$reverse) || ($strand eq "-" && $reverse)  ){
		$startval = $start;
		$endval = $start + $seqlen;
	    }
	    else{
		$seq = reverseComplement($seq);
		$startval = $start + $seqlen;
		$endval = $start;
	    }
	    $keyvalue =  "$tig"."|$startval"."|$endval\n";
	}
    }
    return $seq, $keyvalue;
}


#########END 
